@@ -284,8 +284,6 @@ private:
     friend Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiVertexInputLayout &);
 };
 
-Q_DECLARE_TYPEINFO(QRhiVertexInputLayout, Q_RELOCATABLE_TYPE);
-
 Q_GUI_EXPORT bool operator==(const QRhiVertexInputLayout &a, const QRhiVertexInputLayout &b) noexcept;
 Q_GUI_EXPORT bool operator!=(const QRhiVertexInputLayout &a, const QRhiVertexInputLayout &b) noexcept;
 Q_GUI_EXPORT size_t qHash(const QRhiVertexInputLayout &v, size_t seed = 0) noexcept;
@@ -411,11 +409,35 @@ public:
             StorageImageData simage;
             StorageBufferData sbuf;
         } u;
+
+        template<typename Output>
+        Output serialize(Output dst) const
+        {
+            // must write out exactly LAYOUT_DESC_ENTRIES_PER_BINDING elements here
+            *dst++ = quint32(binding);
+            *dst++ = quint32(stage);
+            *dst++ = quint32(type);
+            *dst++ = quint32(type == QRhiShaderResourceBinding::SampledTexture ? u.stex.count : 1);
+            return dst;
+        }
     };
 
     Data *data() { return &d; }
     const Data *data() const { return &d; }
 
+    static const int LAYOUT_DESC_ENTRIES_PER_BINDING = 4;
+
+    template<typename Output>
+    static void serializeLayoutDescription(const QRhiShaderResourceBinding *first,
+                                           const QRhiShaderResourceBinding *last,
+                                           Output dst)
+    {
+        while (first != last) {
+            dst = first->data()->serialize(dst);
+            ++first;
+        }
+    }
+
 private:
     Data d;
 };
@@ -502,8 +524,6 @@ private:
     QRhiTexture *m_depthTexture = nullptr;
 };
 
-Q_DECLARE_TYPEINFO(QRhiTextureRenderTargetDescription, Q_RELOCATABLE_TYPE);
-
 class Q_GUI_EXPORT QRhiTextureSubresourceUploadDescription
 {
 public:
@@ -518,6 +538,9 @@ public:
     QByteArray data() const { return m_data; }
     void setData(const QByteArray &data) { m_data = data; }
 
+    quint32 dataStride() const { return m_dataStride; }
+    void setDataStride(quint32 stride) { m_dataStride = stride; }
+
     QPoint destinationTopLeft() const { return m_destinationTopLeft; }
     void setDestinationTopLeft(const QPoint &p) { m_destinationTopLeft = p; }
 
@@ -530,6 +553,7 @@ public:
 private:
     QImage m_image;
     QByteArray m_data;
+    quint32 m_dataStride = 0;
     QPoint m_destinationTopLeft;
     QSize m_sourceSize;
     QPoint m_sourceTopLeft;
@@ -581,8 +605,6 @@ private:
     QVarLengthArray<QRhiTextureUploadEntry, 16> m_entries;
 };
 
-Q_DECLARE_TYPEINFO(QRhiTextureUploadDescription, Q_RELOCATABLE_TYPE);
-
 class Q_GUI_EXPORT QRhiTextureCopyDescription
 {
 public:
@@ -749,7 +771,8 @@ public:
         UsedWithGenerateMips = 1 << 6,
         UsedWithLoadStore = 1 << 7,
         UsedAsCompressedAtlas = 1 << 8,
-        ExternalOES = 1 << 9
+        ExternalOES = 1 << 9,
+        ThreeDimensional = 1 << 10
     };
     Q_DECLARE_FLAGS(Flags, Flag)
 
@@ -761,6 +784,7 @@ public:
         R8,
         RG8,
         R16,
+        RG16,
         RED_OR_ALPHA8,
 
         RGBA16F,
@@ -814,6 +838,9 @@ public:
     QSize pixelSize() const { return m_pixelSize; }
     void setPixelSize(const QSize &sz) { m_pixelSize = sz; }
 
+    int depth() const { return m_depth; }
+    void setDepth(int depth) { m_depth = depth; }
+
     Flags flags() const { return m_flags; }
     void setFlags(Flags f) { m_flags = f; }
 
@@ -826,10 +853,11 @@ public:
     virtual void setNativeLayout(int layout);
 
 protected:
-    QRhiTexture(QRhiImplementation *rhi, Format format_, const QSize &pixelSize_,
+    QRhiTexture(QRhiImplementation *rhi, Format format_, const QSize &pixelSize_, int depth_,
                 int sampleCount_, Flags flags_);
     Format m_format;
     QSize m_pixelSize;
+    int m_depth;
     int m_sampleCount;
     Flags m_flags;
 };
@@ -913,6 +941,10 @@ public:
     };
     Q_DECLARE_FLAGS(Flags, Flag)
 
+    struct NativeRenderBuffer {
+        quint64 object;
+    };
+
     QRhiResource::Type resourceType() const override;
 
     Type type() const { return m_type; }
@@ -928,6 +960,7 @@ public:
     void setFlags(Flags h) { m_flags = h; }
 
     virtual bool create() = 0;
+    virtual bool createFrom(NativeRenderBuffer src);
 
     virtual QRhiTexture::Format backingFormat() const = 0;
 
@@ -1022,19 +1055,34 @@ public:
 
     bool isLayoutCompatible(const QRhiShaderResourceBindings *other) const;
 
+    QVector<quint32> serializedLayoutDescription() const { return m_layoutDesc; }
+
     virtual bool create() = 0;
 
+    enum UpdateFlag {
+        BindingsAreSorted = 0x01
+    };
+    Q_DECLARE_FLAGS(UpdateFlags, UpdateFlag)
+
+    virtual void updateResources(UpdateFlags flags = {}) = 0;
+
 protected:
+    static const int BINDING_PREALLOC = 12;
     QRhiShaderResourceBindings(QRhiImplementation *rhi);
-    QVarLengthArray<QRhiShaderResourceBinding, 16> m_bindings;
-    uint m_layoutDescHash = 0;
-    QVarLengthArray<uint, 16 * 3> m_layoutDesc;
+    QVarLengthArray<QRhiShaderResourceBinding, BINDING_PREALLOC> m_bindings;
+    size_t m_layoutDescHash = 0;
+    // Intentionally not using QVLA for m_layoutDesc: clients like Qt Quick are much
+    // better served with an implicitly shared container here, because they will likely
+    // throw this directly into structs serving as cache keys.
+    QVector<quint32> m_layoutDesc;
     friend class QRhiImplementation;
 #ifndef QT_NO_DEBUG_STREAM
     friend Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiShaderResourceBindings &);
 #endif
 };
 
+Q_DECLARE_OPERATORS_FOR_FLAGS(QRhiShaderResourceBindings::UpdateFlags)
+
 #ifndef QT_NO_DEBUG_STREAM
 Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiShaderResourceBindings &);
 #endif
@@ -1413,7 +1461,7 @@ struct Q_GUI_EXPORT QRhiReadbackResult
     QRhiTexture::Format format;
     QSize pixelSize;
     QByteArray data;
-}; // non-movable due to the std::function
+};
 
 struct Q_GUI_EXPORT QRhiBufferReadbackResult
 {
@@ -1528,7 +1576,11 @@ public:
         IntAttributes,
         ScreenSpaceDerivatives,
         ReadBackAnyTextureFormat,
-        PipelineCacheDataLoadSave
+        PipelineCacheDataLoadSave,
+        ImageDataStride,
+        RenderBufferImport,
+        ThreeDimensionalTextures,
+        RenderTo3DTextureSlice
     };
 
     enum BeginFrameFlag {
@@ -1588,6 +1640,11 @@ public:
                             int sampleCount = 1,
                             QRhiTexture::Flags flags = {});
 
+    QRhiTexture *newTexture(QRhiTexture::Format format,
+                            int width, int height, int depth,
+                            int sampleCount = 1,
+                            QRhiTexture::Flags flags = {});
+
     QRhiSampler *newSampler(QRhiSampler::Filter magFilter,
                             QRhiSampler::Filter minFilter,
                             QRhiSampler::Filter mipmapMode,
@@ -1634,8 +1691,7 @@ public:
 
     QRhiProfiler *profiler();
 
-    static const int MAX_LAYERS = 6; // cubemaps only
-    static const int MAX_LEVELS = 16; // a width and/or height of 65536 should be enough for everyone
+    static const int MAX_MIP_LEVELS = 16; // a width and/or height of 65536 should be enough for everyone
 
     void releaseCachedResources();
 
