@@ -50,134 +50,10 @@
 #include <private/qnetworkfile_p.h>
 
 #include <emscripten.h>
-#include <emscripten/bind.h>
-#include <emscripten/val.h>
+#include <emscripten/fetch.h>
 
 QT_BEGIN_NAMESPACE
 
-using namespace emscripten;
-
-static void q_requestErrorCallback(val event)
-{
-    val xhr = event["target"];
-
-    quintptr func = xhr["data-handler"].as<quintptr>();
-    QNetworkReplyWasmImplPrivate *reply = reinterpret_cast<QNetworkReplyWasmImplPrivate*>(func);
-    Q_ASSERT(reply);
-
-    int statusCode = xhr["status"].as<int>();
-
-    QString reasonStr = QString::fromStdString(xhr["statusText"].as<std::string>());
-
-    reply->setReplyAttributes(func, statusCode, reasonStr);
-
-    if (statusCode >= 400 && !reasonStr.isEmpty())
-        reply->emitReplyError(reply->statusCodeFromHttp(statusCode, reply->request.url()), reasonStr);
-}
-
-static void q_progressCallback(val event)
-{
-    val xhr = event["target"];
-
-    QNetworkReplyWasmImplPrivate *reply =
-            reinterpret_cast<QNetworkReplyWasmImplPrivate*>(xhr["data-handler"].as<quintptr>());
-    Q_ASSERT(reply);
-
-    if (xhr["lengthComputable"].as<bool>() && xhr["status"].as<int>() < 400)
-        reply->emitDataReadProgress(xhr["loaded"].as<qint64>(), xhr["total"].as<qint64>());
-
-}
-
-static void q_loadCallback(val event)
-{
-    val xhr = event["target"];
-
-    QNetworkReplyWasmImplPrivate *reply =
-            reinterpret_cast<QNetworkReplyWasmImplPrivate*>(xhr["data-handler"].as<quintptr>());
-    Q_ASSERT(reply);
-
-    int status = xhr["status"].as<int>();
-    if (status >= 300) {
-        q_requestErrorCallback(event);
-        return;
-    }
-    QString statusText = QString::fromStdString(xhr["statusText"].as<std::string>());
-    if (status == 200 || status == 203) {
-        QString responseString;
-        const std::string responseType = xhr["responseType"].as<std::string>();
-        if (responseType.length() == 0 || responseType == "document" || responseType == "text") {
-            responseString = QString::fromStdWString(xhr["responseText"].as<std::wstring>());
-        } else if (responseType == "json") {
-            responseString =
-                    QString::fromStdWString(val::global("JSON").call<std::wstring>("stringify", xhr["response"]));
-        } else if (responseType == "arraybuffer" || responseType == "blob") {
-            // handle this data in the FileReader, triggered by the call to readAsArrayBuffer
-            val reader = val::global("FileReader").new_();
-            reader.set("onload", val::module_property("QNetworkReplyWasmImplPrivate_readBinary"));
-            reader.set("data-handler", xhr["data-handler"]);
-            reader.call<void>("readAsArrayBuffer", xhr["response"]);
-        }
-
-        int readyState = xhr["readyState"].as<int>();
-
-        if (readyState == 4) { // done
-            reply->setReplyAttributes(xhr["data-handler"].as<quintptr>(), status, statusText);
-            if (!responseString.isEmpty())
-                reply->dataReceived(responseString.toUtf8(), responseString.size());
-        }
-    }
-    if (status >= 400 && !statusText.isEmpty())
-        reply->emitReplyError(reply->statusCodeFromHttp(status, reply->request.url()), statusText);
-}
-
-static void q_responseHeadersCallback(val event)
-{
-    val xhr = event["target"];
-
-    if (xhr["readyState"].as<int>() == 2) { // HEADERS_RECEIVED
-        std::string responseHeaders = xhr.call<std::string>("getAllResponseHeaders");
-        if (!responseHeaders.empty()) {
-            QNetworkReplyWasmImplPrivate *reply =
-                    reinterpret_cast<QNetworkReplyWasmImplPrivate*>(xhr["data-handler"].as<quintptr>());
-            Q_ASSERT(reply);
-
-            reply->headersReceived(QString::fromStdString(responseHeaders));
-        }
-    }
-}
-
-static void q_readBinary(val event)
-{
-    val fileReader = event["target"];
-
-    QNetworkReplyWasmImplPrivate *reply =
-            reinterpret_cast<QNetworkReplyWasmImplPrivate*>(fileReader["data-handler"].as<quintptr>());
-    Q_ASSERT(reply);
-
-    // Set up source typed array
-    val result = fileReader["result"]; // ArrayBuffer
-    val Uint8Array = val::global("Uint8Array");
-    val sourceTypedArray = Uint8Array.new_(result);
-
-    // Allocate and set up destination typed array
-    const quintptr size = result["byteLength"].as<quintptr>();
-    QByteArray buffer(size, Qt::Uninitialized);
-
-    val destinationTypedArray = Uint8Array.new_(val::module_property("HEAPU8")["buffer"],
-                                                            reinterpret_cast<quintptr>(buffer.data()), size);
-    destinationTypedArray.call<void>("set", sourceTypedArray);
-    reply->dataReceived(buffer, buffer.size());
-}
-
-
-EMSCRIPTEN_BINDINGS(network_module) {
-    function("QNetworkReplyWasmImplPrivate_requestErrorCallback", q_requestErrorCallback);
-    function("QNetworkReplyWasmImplPrivate_progressCallback", q_progressCallback);
-    function("QNetworkReplyWasmImplPrivate_loadCallback", q_loadCallback);
-    function("QNetworkReplyWasmImplPrivate_responseHeadersCallback", q_responseHeadersCallback);
-    function("QNetworkReplyWasmImplPrivate_readBinary", q_readBinary);
-}
-
 QNetworkReplyWasmImplPrivate::QNetworkReplyWasmImplPrivate()
     : QNetworkReplyPrivate()
     , managerPrivate(0)
@@ -185,6 +61,7 @@ QNetworkReplyWasmImplPrivate::QNetworkReplyWasmImplPrivate()
     , downloadBufferCurrentSize(0)
     , totalDownloadSize(0)
     , percentFinished(0)
+    , m_fetch(0)
 {
 }
 
@@ -192,17 +69,20 @@ QNetworkReplyWasmImplPrivate::~QNetworkReplyWasmImplPrivate()
 {
 }
 
-QNetworkReplyWasmImpl::~QNetworkReplyWasmImpl()
-{
-}
-
 QNetworkReplyWasmImpl::QNetworkReplyWasmImpl(QObject *parent)
     : QNetworkReply(*new QNetworkReplyWasmImplPrivate(), parent)
+{
+    Q_D( QNetworkReplyWasmImpl);
+    d->state = QNetworkReplyPrivate::Idle;
+}
+
+QNetworkReplyWasmImpl::~QNetworkReplyWasmImpl()
 {
 }
 
 QByteArray QNetworkReplyWasmImpl::methodName() const
 {
+    const Q_D( QNetworkReplyWasmImpl);
     switch (operation()) {
     case QNetworkAccessManager::HeadOperation:
         return "HEAD";
@@ -214,6 +94,8 @@ QByteArray QNetworkReplyWasmImpl::methodName() const
         return "POST";
     case QNetworkAccessManager::DeleteOperation:
         return "DELETE";
+    case QNetworkAccessManager::CustomOperation:
+        return d->request.attribute(QNetworkRequest::CustomVerbAttribute).toByteArray();
     default:
         break;
     }
@@ -222,17 +104,21 @@ QByteArray QNetworkReplyWasmImpl::methodName() const
 
 void QNetworkReplyWasmImpl::close()
 {
+    QNetworkReply::close();
     setFinished(true);
     emit finished();
-
-    QNetworkReply::close();
 }
 
 void QNetworkReplyWasmImpl::abort()
 {
-    Q_D(const QNetworkReplyWasmImpl);
-    d->doAbort();
+    Q_D(QNetworkReplyWasmImpl);
+    if (d->state == QNetworkReplyPrivate::Finished || d->state == QNetworkReplyPrivate::Aborted)
+        return;
 
+    d->state = QNetworkReplyPrivate::Aborted;
+    d->m_fetch->userData = nullptr;
+
+    d->emitReplyError(QNetworkReply::OperationCanceledError, QStringLiteral("Operation canceled"));
     close();
 }
 
@@ -240,9 +126,6 @@ qint64 QNetworkReplyWasmImpl::bytesAvailable() const
 {
     Q_D(const QNetworkReplyWasmImpl);
 
-    if (!d->isFinished)
-        return QNetworkReply::bytesAvailable();
-
     return QNetworkReply::bytesAvailable() + d->downloadBufferCurrentSize - d->downloadBufferReadPosition;
 }
 
@@ -264,7 +147,7 @@ qint64 QNetworkReplyWasmImpl::readData(char *data, qint64 maxlen)
     Q_D(QNetworkReplyWasmImpl);
 
     qint64 howMuch = qMin(maxlen, (d->downloadBuffer.size() - d->downloadBufferReadPosition));
-    memcpy(data, d->downloadBuffer.constData(), howMuch);
+    memcpy(data, d->downloadBuffer.constData() + d->downloadBufferReadPosition, howMuch);
     d->downloadBufferReadPosition += howMuch;
 
     return howMuch;
@@ -313,9 +196,8 @@ void QNetworkReplyWasmImplPrivate::setReplyAttributes(quintptr data, int statusC
         handler->q_func()->setAttribute(QNetworkRequest::HttpReasonPhraseAttribute, statusReason);
 }
 
-void QNetworkReplyWasmImplPrivate::doAbort() const
-{
-    m_xhr.call<void>("abort");
+constexpr int getArraySize (int factor) {
+    return 2 * factor + 1;
 }
 
 void QNetworkReplyWasmImplPrivate::doSendRequest()
@@ -323,69 +205,77 @@ void QNetworkReplyWasmImplPrivate::doSendRequest()
     Q_Q(QNetworkReplyWasmImpl);
     totalDownloadSize = 0;
 
-    m_xhr = val::global("XMLHttpRequest").new_();
-    std::string verb = q->methodName().toStdString();
+    emscripten_fetch_attr_t attr;
+    emscripten_fetch_attr_init(&attr);
+    strcpy(attr.requestMethod, q->methodName().constData());
 
-    QString extraDataString;
+    QList<QByteArray> headersData = request.rawHeaderList();
+    int arrayLength = getArraySize(headersData.count());
+    const char* customHeaders[arrayLength];
 
-    m_xhr.call<void>("open", verb, request.url().toString().toStdString());
-
-    m_xhr.set("onerror", val::module_property("QNetworkReplyWasmImplPrivate_requestErrorCallback"));
-    m_xhr.set("onload", val::module_property("QNetworkReplyWasmImplPrivate_loadCallback"));
-    m_xhr.set("onprogress", val::module_property("QNetworkReplyWasmImplPrivate_progressCallback"));
-    m_xhr.set("onreadystatechange", val::module_property("QNetworkReplyWasmImplPrivate_responseHeadersCallback"));
-
-    m_xhr.set("data-handler", val(quintptr(reinterpret_cast<void *>(this))));
-
-    QByteArray contentType = request.rawHeader("Content-Type");
-
-    // handle extra data
-    val dataToSend = val::null();
-    QByteArray extraData;
-
-    if (outgoingData) // data from post request
-        extraData = outgoingData->readAll();
-
-    if (contentType.contains("text") ||
-            contentType.contains("json") ||
-            contentType.contains("form")) {
-        if (extraData.size() > 0)
-            extraDataString.fromUtf8(extraData);
+    if (headersData.count() > 0) {
+        int i = 0;
+        for (int j = 0; j < headersData.count(); j++) {
+            customHeaders[i] = headersData[j].constData();
+            i += 1;
+            customHeaders[i] = request.rawHeader(headersData[j]).constData();
+            i += 1;
+        }
+        customHeaders[i] = nullptr;
+        attr.requestHeaders = customHeaders;
     }
-    if (contentType.contains("json")) {
-        if (!extraDataString.isEmpty()) {
-            m_xhr.set("responseType", val("json"));
-            dataToSend = val(extraDataString.toStdString());
+
+    if (outgoingData) { // data from post request
+        // handle extra data
+        requestData = outgoingData->readAll(); // is there a size restriction here?
+        if (!requestData.isEmpty()) {
+            attr.requestData = requestData.data();
+            attr.requestDataSize = requestData.size();
         }
     }
-    if (contentType.contains("form")) { //construct form data
-        if (!extraDataString.isEmpty()) {
-            val formData = val::global("FormData").new_();
-            QStringList formList = extraDataString.split('&');
 
-            for (auto formEntry : formList) {
-                formData.call<void>("append", formEntry.split('=')[0].toStdString(), formEntry.split('=')[1].toStdString());
-            }
-            dataToSend = formData;
-        }
+    // username & password
+    if (!request.url().userInfo().isEmpty()) {
+        attr.userName = request.url().userName().toUtf8();
+        attr.password = request.url().password().toUtf8();
     }
-    // set request headers
-    for (auto header : request.rawHeaderList()) {
-        m_xhr.call<void>("setRequestHeader", header.toStdString(), request.rawHeader(header).toStdString());
+
+    attr.attributes = EMSCRIPTEN_FETCH_LOAD_TO_MEMORY;
+
+    QNetworkRequest::CacheLoadControl CacheLoadControlAttribute =
+        (QNetworkRequest::CacheLoadControl)request.attribute(QNetworkRequest::CacheLoadControlAttribute, QNetworkRequest::PreferNetwork).toInt();
+
+    if (CacheLoadControlAttribute == QNetworkRequest::AlwaysCache) {
+        attr.attributes += EMSCRIPTEN_FETCH_NO_DOWNLOAD;
     }
-     m_xhr.call<void>("send", dataToSend);
+    if (CacheLoadControlAttribute == QNetworkRequest::PreferCache) {
+         attr.attributes += EMSCRIPTEN_FETCH_APPEND;
+    }
+
+    if (CacheLoadControlAttribute == QNetworkRequest::AlwaysNetwork ||
+            request.attribute(QNetworkRequest::CacheSaveControlAttribute, false).toBool()) {
+        attr.attributes -= EMSCRIPTEN_FETCH_PERSIST_FILE;
+    }
+
+    attr.onsuccess = QNetworkReplyWasmImplPrivate::downloadSucceeded;
+    attr.onerror = QNetworkReplyWasmImplPrivate::downloadFailed;
+    attr.onprogress = QNetworkReplyWasmImplPrivate::downloadProgress;
+    attr.onreadystatechange = QNetworkReplyWasmImplPrivate::stateChange;
+    attr.timeoutMSecs = request.transferTimeout();
+    attr.userData = reinterpret_cast<void *>(this);
+
+    QString dPath = QStringLiteral("/home/web_user/") + request.url().fileName();
+    attr.destinationPath = dPath.toUtf8();
+
+    m_fetch = emscripten_fetch(&attr, request.url().toString().toUtf8());
 }
 
 void QNetworkReplyWasmImplPrivate::emitReplyError(QNetworkReply::NetworkError errorCode, const QString &errorString)
 {
-    Q_UNUSED(errorCode)
     Q_Q(QNetworkReplyWasmImpl);
 
     q->setError(errorCode, errorString);
-    emit q->error(errorCode);
-
-    q->setFinished(true);
-    emit q->finished();
+    emit q->errorOccurred(errorCode);
 }
 
 void QNetworkReplyWasmImplPrivate::emitDataReadProgress(qint64 bytesReceived, qint64 bytesTotal)
@@ -417,10 +307,13 @@ void QNetworkReplyWasmImplPrivate::dataReceived(const QByteArray &buffer, int bu
 
     downloadBuffer.append(buffer, bufferSize);
 
+    emit q->readyRead();
+
     if (downloadBufferCurrentSize == totalDownloadSize) {
-         q->setFinished(true);
-         emit q->finished();
-     }
+        q->setFinished(true);
+        emit q->readChannelFinished();
+        emit q->finished();
+    }
 }
 
 //taken from qnetworkrequest.cpp
@@ -463,25 +356,28 @@ static int parseHeaderName(const QByteArray &headerName)
 }
 
 
-void QNetworkReplyWasmImplPrivate::headersReceived(const QString &bufferString)
+void QNetworkReplyWasmImplPrivate::headersReceived(const QByteArray &buffer)
 {
     Q_Q(QNetworkReplyWasmImpl);
 
-    if (!bufferString.isEmpty()) {
-        QStringList headers = bufferString.split(QString::fromUtf8("\r\n"), QString::SkipEmptyParts);
+    if (!buffer.isEmpty()) {
+        QList<QByteArray> headers = buffer.split('\n');
 
         for (int i = 0; i < headers.size(); i++) {
-            QString headerName = headers.at(i).split(QString::fromUtf8(": ")).at(0);
-            QString headersValue = headers.at(i).split(QString::fromUtf8(": ")).at(1);
-            if (headerName.isEmpty() || headersValue.isEmpty())
-                continue;
+            if (headers.at(i).contains(':')) { // headers include final \x00, so skip
+                QByteArray headerName = headers.at(i).split(':').at(0).trimmed();
+                QByteArray headersValue = headers.at(i).split(':').at(1).trimmed();
 
-            int headerIndex = parseHeaderName(headerName.toLocal8Bit());
+                if (headerName.isEmpty() || headersValue.isEmpty())
+                    continue;
 
-            if (headerIndex == -1)
-                q->setRawHeader(headerName.toLocal8Bit(), headersValue.toLocal8Bit());
-            else
-                q->setHeader(static_cast<QNetworkRequest::KnownHeaders>(headerIndex), (QVariant)headersValue);
+                int headerIndex = parseHeaderName(headerName);
+
+                if (headerIndex == -1)
+                    q->setRawHeader(headerName, headersValue);
+                else
+                    q->setHeader(static_cast<QNetworkRequest::KnownHeaders>(headerIndex), (QVariant)headersValue);
+            }
         }
     }
     emit q->metaDataChanged();
@@ -547,6 +443,72 @@ void QNetworkReplyWasmImplPrivate::_q_bufferOutgoingData()
     }
 }
 
+void QNetworkReplyWasmImplPrivate::downloadSucceeded(emscripten_fetch_t *fetch)
+{
+    auto reply = reinterpret_cast<QNetworkReplyWasmImplPrivate*>(fetch->userData);
+    if (reply) {
+        if (reply->state != QNetworkReplyPrivate::Aborted) {
+            QByteArray buffer(fetch->data, fetch->numBytes);
+            reply->dataReceived(buffer, buffer.size());
+        }
+        reply->m_fetch = nullptr;
+    }
+    emscripten_fetch_close(fetch);
+}
+
+void QNetworkReplyWasmImplPrivate::setStatusCode(int status, const QByteArray &statusText)
+{
+    Q_Q(QNetworkReplyWasmImpl);
+    q->setAttribute(QNetworkRequest::HttpStatusCodeAttribute, status);
+    q->setAttribute(QNetworkRequest::HttpReasonPhraseAttribute, statusText);
+}
+
+void QNetworkReplyWasmImplPrivate::stateChange(emscripten_fetch_t *fetch)
+{
+    auto reply = reinterpret_cast<QNetworkReplyWasmImplPrivate*>(fetch->userData);
+    if (reply && reply->state != QNetworkReplyPrivate::Aborted) {
+        if (fetch->readyState == /*HEADERS_RECEIVED*/ 2) {
+            size_t headerLength = emscripten_fetch_get_response_headers_length(fetch);
+            QByteArray str(headerLength, Qt::Uninitialized);
+            emscripten_fetch_get_response_headers(fetch, str.data(), str.size());
+            reply->headersReceived(str);
+        }
+    }
+}
+
+void QNetworkReplyWasmImplPrivate::downloadProgress(emscripten_fetch_t *fetch)
+{
+    auto reply = reinterpret_cast<QNetworkReplyWasmImplPrivate*>(fetch->userData);
+    if (reply && reply->state != QNetworkReplyPrivate::Aborted) {
+        if (fetch->status < 400) {
+            uint64_t bytes = fetch->dataOffset + fetch->numBytes;
+            uint64_t tBytes = fetch->totalBytes; // totalBytes can be 0 if server not reporting content length
+            if (tBytes == 0)
+                tBytes = bytes;
+            reply->emitDataReadProgress(bytes, tBytes);
+        }
+    }
+}
+
+void QNetworkReplyWasmImplPrivate::downloadFailed(emscripten_fetch_t *fetch)
+{
+    auto reply = reinterpret_cast<QNetworkReplyWasmImplPrivate*>(fetch->userData);
+    if (reply) {
+        if (reply->state != QNetworkReplyPrivate::Aborted) {
+            QString reasonStr;
+            if (fetch->status > 600)
+                reasonStr = QStringLiteral("Operation canceled");
+            else
+                reasonStr = QString::fromUtf8(fetch->statusText);
+            QByteArray statusText(fetch->statusText);
+            reply->setStatusCode(fetch->status, statusText);
+            reply->emitReplyError(reply->statusCodeFromHttp(fetch->status, reply->request.url()), reasonStr);
+        }
+        reply->m_fetch = nullptr;
+    }
+    emscripten_fetch_close(fetch);
+}
+
 //taken from qhttpthreaddelegate.cpp
 QNetworkReply::NetworkError QNetworkReplyWasmImplPrivate::statusCodeFromHttp(int httpStatusCode, const QUrl &url)
 {
@@ -601,6 +563,9 @@ QNetworkReply::NetworkError QNetworkReplyWasmImplPrivate::statusCodeFromHttp(int
         code = QNetworkReply::ServiceUnavailableError;
         break;
 
+    case 65535: //emscripten reply when aborted
+        code =  QNetworkReply::OperationCanceledError;
+        break;
     default:
         if (httpStatusCode > 500) {
             // some kind of server error
